<?php

/**
 * Allows for the automatic saving of attributes in related HAS_ONE type relations.
 *
 * @author Luke Jurgs
 * @version 0.0.4-2013-02-03
 *
 * The following are virtual attributes:
 * @property CActiveRecord $owner.
 * @property array $compositeAttributes
 * @property array $relatedModels
 *
 * @method ActiveRecord getOwner
 *
 * Requires that the owner CActiveRecord defines the method compositeAttributes() that returns an array:
 * array(
 *  'myAttribute' => array('RelatedRecord', 'relatedRecordAttribute'),
 *  'myAttribute2' => array('RelatedRecord2', 'relatedRecordAttribute'),
 * );
 */
class ECompositeBehavior extends CActiveRecordBehavior
{
	private $_compositeAttributes;
	private $_relatedModels;

	/**
	 * Overrides the parent implementation to allow getting values for attributes described in the compositeAttributes
	 * method from the owner CActiveRecord.
	 * @param string $name the name of the attribute.
	 * @return mixed|null
	 */
	public function __get($name)
	{
		$compositeAttributes = $this->getCompositeAttributes();
		if (isset($compositeAttributes[$name]))
		{
			if (array_key_exists('value', $compositeAttributes[$name]))
			{
				return $compositeAttributes[$name]['value'];
			}
			elseif (null === ($relation = $this->getOwner()->getRelated($compositeAttributes[$name][0])))
			{
				return null;
			}
			else
			{
				return $relation->$compositeAttributes[$name][1];
			}
		}
		else
		{
			return parent::__get($name);
		}
	}

	/**
	 * Overrides the parent implementation to allow setting values for attributes described in the compositeAttributes
	 * method from the owner CActiveRecord.
	 * @param string $name the name of the attribute.
	 * @param mixed $value value to populate the attribute with.
	 * @return mixed
	 */
	public function __set($name, $value)
	{
		$compositeAttributes = $this->getCompositeAttributes();
		if (isset($compositeAttributes[$name]))
		{
			/**
			 * need to invalidate the related model cache because if a save is made and another update occurs any
			 * subsequent save will fail to correctly account for any newly set attributes.
			 */
			$this->_relatedModels = null;
			return $this->_compositeAttributes[$name]['value'] = $value;
		}
		else
		{
			return parent::__set($name, $value);
		}
	}

	/**
	 * Overrides the parent implementation to check if it's possible to get a value from a composite attribute.
	 * @param string $name the name of the attribute or property.
	 * @return bool
	 */
	public function canGetProperty($name)
	{
		$compositeAttributes = $this->getCompositeAttributes();
		if (isset($compositeAttributes[$name]))
		{
			return true;
		}
		else
		{
			return parent::canGetProperty($name);
		}
	}

	/**
	 * Overrides the parent implementation to check if it's possible to set a value to a composite attribute.
	 * @param string $name the name of the attribute or property.
	 * @return bool
	 */
	public function canSetProperty($name)
	{
		$compositeAttributes = $this->getCompositeAttributes();
		if (isset($compositeAttributes[$name]))
		{
			return true;
		}
		else
		{
			return parent::canSetProperty($name);
		}
	}

	/**
	 * Returns all composite attribute values.
	 * @param array $names names of attributes whose value needs to be returned.
	 * @return array attribute values indexed by attribute names.
	 */
	public function getAttributes($names = null)
	{
		$attributes = array();
		foreach ($this->getCompositeAttributes() as $compositeAttributeName => $compositeAttribute)
		{
			if (!is_array($names) || in_array($compositeAttributeName, $names))
			{
				$attributes[$compositeAttributeName] = $this->$compositeAttribute[1];
			}
		}
		return $attributes;
	}

	/**
	 * Get the attribute labels for the composite model attributes.
	 * @return array
	 */
	public function attributeLabels()
	{
		$ownerRelations = $this->getOwner()->relations();
		$compositeAttributeLabels = array();
		$relatedModels = array();
		foreach ($this->getCompositeAttributes() as $compositeAttributeName => $compositeAttribute)
		{
			$relationName = $compositeAttribute[0];
			if (!isset($relatedModels[$relationName]))
			{
				if ((null !== $this->_relatedModels) && (isset($this->_relatedModels[$relationName])))
				{ //if we already have an instance of the related record
					$relatedModels[$relationName] = $this->_relatedModels[$relationName];
				}
				else
				{ //else instantiate
					$relatedModels[$relationName] = new $ownerRelations[$relationName][1];
				}
			}
			$compositeAttributeLabels[$compositeAttributeName] = $relatedModels[$relationName]->getAttributeLabel($compositeAttribute[1]);
		}
		return $compositeAttributeLabels;
	}

	public function afterConstruct($event)
	{
		$this->addValidator();
	}

	public function afterFind($event)
	{
		$this->addValidator();
	}

	/**
	 * Runs validation against the related record attributes.
	 * @param CEvent $event
	 */
	public function afterValidate($event)
	{
		/** @var CActiveRecord $model */
		foreach ($this->getRelatedModels() as $model)
		{
			if (!$this->getIsEmpty($model, true))
			{
				$attributes = $model->attributes;
				$tableName = $this->getOwner()->tableName();
				foreach ($model->tableSchema->foreignKeys as $key => $foreignKey)
				{ //do not validate the the foreign key attributes since they are set at save
					if (0 === strcmp($tableName, $foreignKey[0]))
					{
						unset($attributes[$key]);
					}
				}
				if (!$model->validate(array_keys($attributes)))
				{ //if validation fails on anything, add the error and save will fail
					$this->getOwner()->addErrors($model->getErrors());
				}
			}
		}
	}

	/**
	 * This actually pushes the data into the composite attributes.
	 * - if attributes are set and the related record exists - update and save.
	 * - if attributes are set and related record does not exist - create new model and save.
	 * - if attributes are set but empty and related record exists - delete.
	 * - else do nothing.
	 */
	public function afterSave()
	{
		//now save or delete related data based on it's existence
		/** @var CActiveRecord $model */
		foreach ($this->getRelatedModels() as $model)
		{
			if ($this->getIsEmpty($model, true))
			{
				if (!$model->isNewRecord)
				{
					$model->delete();
				}
			}
			else
			{
				if ($model->isNewRecord)
				{
					$tableName = $this->getOwner()->tableName();
					foreach ($model->getTableSchema()->foreignKeys as $key => $foreignKey)
					{ //populate the foreign keys before saving a new record
						if (0 === strcmp($tableName, $foreignKey[0]))
						{
							$model->$key = $this->getOwner()->$foreignKey[1];
						}
					}
				}
				$model->save(false);
			}
		}
	}

	/**
	 * Get the composite attributes descriptor array from the owner ActiveModel.
	 * @return array
	 */
	protected function getCompositeAttributes()
	{
		if (null === $this->_compositeAttributes)
		{
			$this->_compositeAttributes = $this->getOwner()->compositeAttributes();
		}
		return $this->_compositeAttributes;
	}

	/**
	 * Get the related models, populated with any assigned composite attributes.
	 * Only returns models where something was set.
	 * @return ActiveRecord[] - empty array if no attributes were assigned to any composite attributes.
	 */
	protected function getRelatedModels()
	{
		if (null === $this->_relatedModels)
		{
			$compositeAttributes = $this->getCompositeAttributes();
			$ownerRelations = $this->getOwner()->relations();
			$relatedModels = array();
			foreach ($compositeAttributes as $compositeAttribute)
			{
				$relationName = $compositeAttribute[0];
				if (array_key_exists('value', $compositeAttribute))
				{ //only proceed if something was assigned. (if nothing was assigned at all don't want to interact with this model)
					if (!isset($relatedModels[$relationName]))
					{ //a relation might be listed many times, only instantiate/retrieve once
						/** @var CActiveRecord $relation */
						if (null === ($relation = $this->getOwner()->getRelated($relationName)))
						{ //if there is no such related model create an instance
							$relation = new $ownerRelations[$relationName][1];
						}
						$relatedModels[$relationName] = $relation;
					}
					$relatedModels[$relationName]->$compositeAttribute[1] = $compositeAttribute['value'];
				}
			}
			$this->_relatedModels = $relatedModels;
		}
		return $this->_relatedModels;
	}

	/**
	 * Attach a safe validator to the owner for each attribute.
	 */
	protected function addValidator()
	{
		$validatorList = $this->getOwner()->getValidatorList();
		$attributes = array_keys($this->getCompositeAttributes());
		$validatorList->add(CValidator::createValidator('safe', $this->getOwner(), $attributes));
	}

	/**
	 * Determines if the column attributes in the supplied record are all empty.
	 * @param CActiveRecord $model
	 * @param bool $excludePrimaryKey exclude the primary key from the operation.
	 * @return bool this record is empty.
	 */
	private function getIsEmpty($model, $excludePrimaryKey = false)
	{
		$attributes = $model->attributes;
		if ($excludePrimaryKey)
		{
			$primaryKey = $model->tableSchema->primaryKey;
			if (is_string($primaryKey))
			{
				$primaryKey = array($primaryKey);
			}
			$primaryKey = array_flip($primaryKey);
			$attributes = array_diff_key($attributes, $primaryKey);
		}
		foreach ($attributes as $attribute)
		{
			if (is_string($attribute) && 0 !== strlen($attribute))
			{
				return false;
			}
		}
		return true;
	}
}